/*
 *      b t r e e  _ c
 *
 *  This contains routines that are specific
 *  to the implementation of binary trees
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include "liblistx.h"


/*======================================================= BTREE_ADDNODE */
static
void    BTREE_AddNode   (list_t list, _btree_t newnode, _btree_t listnode)
/*      ~~~~~~~~~~~~
 *  This is a routine that tries to add a node starting at the
 *  current point in the tree.  It could be written more compactly
 *  by using recursion, but this adds significant extra runtime
 *  overhead.
 *----------------------------------------------------------------------*/
{
    /*
     *  If the list is empty, then
     *  this is a special case.
     */
    if (listnode == NULL)
    {
        ((_btreeh_t)list)->nodes = newnode;
        return;
    }


    /*
     *  List not empty.
     *  This is the part where recursion
     *  could help make more compact code.
     */
    for ( ; ;)
    {
        /*
         *  Handle less than case
         */
        while (_LIST_Compare (list, (node_t)newnode, (node_t)listnode) < 0)
        {
            if (listnode->left == NULL)
            {
                listnode->left = newnode;
                newnode->up = listnode;
                return;
            }
            listnode = listnode->left;
        }
        /*
         *  Handle greater than case
         */
        while (_LIST_Compare (list, (node_t)newnode, (node_t)listnode) >= 0)
        {
            if (listnode->right == NULL)
            {
                listnode->right = newnode;
                newnode->up = listnode;
                return;
            }
            listnode = listnode->right;
        }
    }
}


/*========================================================= BTREE_AFTER */
static
int     BTREE_After     (list_t list, node_t newnode, node_t oldnode)
/*      ~~~~~~~~~~~
 *  Add a node after the given node.
 *-----------------------------------------------------------------------*/
#define _newnode  ((_btree_t)newnode)
#define _oldnode  ((_btree_t)oldnode)
{
    LIBASSERT ((list->nodes == NULL && oldnode == NULL)
          ||(list->nodes != NULL && oldnode != NULL) );

    LIBDEBUG ("BTREE_After", \
           "'newnode' appears to already be linked in to a list", \
           _newnode->up == NULL && _newnode->left == NULL && _newnode->right == NULL)

    BTREE_AddNode (list, _newnode, _oldnode);

    return (LIST_ERROR_NONE);
}
#undef _newnode
#undef _oldnode


/*=========================================================== BTREE_ADD */
static
int     BTREE_Add       (list_t list, node_t node)
/*      ~~~~~~~~~
 *  Add a node at the correct point.
 *----------------------------------------------------------------------*/
#define _list ((_btreeh_t)list)
#define _node ((_btree_t)node)
{
    LIBDEBUG ("BTREE_Add", \
           "'node' appears to already be linked in to a list", \
           _node->up == NULL && _node->left == NULL && _node->right == NULL)

    BTREE_AddNode(list, _node, _list->nodes);

    return (LIST_ERROR_NONE);
}
#undef _list
#undef _node



/*========================================================================
 *
 *  This is the list of vectored functions
 *  that are used to support this list type.
 *
 *------------------------------------------------------------------------*/

static struct _listmaster master =
        {
        &_LIST_ListCount,
        &_LIST_BtreeFirst,
        &_LIST_BtreeLast,
        &_LIST_BtreeNext,
        &_LIST_BtreePrevious,
        &_LIST_CloneList,
        &_LIST_Destroy,
        &_LIST_NewNode,
        &BTREE_Add,
        &BTREE_After,
        &_LIST_BtreeRemove,
        &_LIST_BtreeFind,
        LIST_TYPE_BTREE,
        &_LISTbtree,
#ifdef DEBUG
        "Binary Tree"
#endif /* DEBUG */
        };

/*
 *  This is the default settings for a new
 *  object of this type.
 */
struct _listclass _LISTbtree =
        {
        &master,                    /* Pointer to master class data */
        sizeof(struct _btreeh),     /* Size of the list header */
        sizeof(struct _btree),      /* Size of the node header */
        0,                          /* Size of user node data */
        0,                          /* Clone count */
        NULL,                       /* User node Initialise */
        NULL,                       /* User node Destroy */
        NULL,                       /* user node Compare */
        };


