/*
 *      E m b e d . c
 *
 *	This file contains routines associated with the
 *	handling of embedded lists that are not suitable
 *	for inclusion into a RLL (because they update
 *	global areas).
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  20 June 96  DJW   - First version
 *
 *	10 July 96  DJW   - Completely re-written as part of redoing embedding.
 */

#include <liblist.h>
#include "liblistx.h"


/*=========================================================== EMBED_CLONE */
static
list_t  EMBED_Clone (list_t oldlist)
/*      ~~~~~~~~~~~
 *
 *  To clone an embedded list, we need to clone
 *  both the user lists and the actual embedded
 *  ones.  They all then need to be linked via
 *  a new 'embed' structure.
 *------------------------------------------------------------------------*/
{
	_embedh_t	oldembed;
	_embedh_t	newouter;
	_embedh_t	newinner;

	for (newouter = NULL, oldembed = _LIST_EmbedOuter(oldlist) ;
				oldembed != NULL;
						oldembed = oldembed->inner)
	{
		if ((newinner = (_embedh_t)_LIST_CloneList ((list_t)oldembed)) == NULL)
		{
			/*
			 *	We should tidy up, but this is rather hard!
			 */
			return NULL;
		}
		if ((newinner->self = _LIST_CloneList (oldembed->self)) == NULL)
		{
			_LIST_UnClone ((list_t)newinner);
			/*
			 *	We should tidy up further, but this is rather hard!
			 */
			return NULL;
		}
		/*
		 *	Now link list levels
		 */
		if (newouter != NULL)
		{
			newinner->outer = newouter;
			newouter->inner = newinner;
		}
		newouter = newinner;
	}
	return (list_t)_LIST_EmbedOuter((list_t)newouter);
}


/*========================================================== EMBED_DESTROY */
static
int     EMBED_Destroy (list_t * plist)
/*      ~~~~~~~~~~~~~
 *  The destroy function for an embedded list
 *  has rather more than usual to do as it has
 *  to get rid off all the extra data structures
 *  created as part of the embedding process.
 *-------------------------------------------------------------------------*/
{
    _embedh_t   embedlist;
    _embedh_t   innerlist;
    node_t      node;

    /*
     * Start by removing all data nodes
     * (from all instances of lists).
     */
    while ((node = LIST_First(*plist)) != NULL)
    {
        (void)LIST_Remove(*plist,node);
        (void)LIST_FreeNode(*plist,&node);
    }
    /*
	 *	Now work down the lists,
	 *  freeing them all in turn.
	 */
	for (innerlist=NULL,embedlist = _LIST_EmbedOuter(*plist) ; 
				embedlist != NULL ; 
						embedlist = innerlist)
	{
		(void)LIST_Destroy (&embedlist->self);
	
		innerlist = embedlist->inner;
		_LIST_UnClone ((list_t)embedlist);
	}
    return LIST_ERROR_NONE;
}



/*========================================================= LIST_EMBED */
extern
list_t  LIST_Embed  (listclass_t newclass,
                     list_t *    poldlist,
                     int  (*compare)(node_t, node_t))
/*
 *  This does the processing to embed one list in another.
 *  It is the heart of creating the embedded lists, and
 *  is by far the most complicated part to handle.
 *
 *  Embedding requires the following basic actions to be
 *  carried out:
 *   a)  Change the list details attached to the users current
 *       list handle so that the master record vectors point
 *       to the "Embed" variants.
 *   b)  Create a new list which has the master record vectors
 *       pointing to the "Embed" variants.
 *   c)  Use the "oldlist" to create the innerlist details.
 *   d)  Create a new list holding the outerlist details.
 *   e)  Create an "embed" structure to link these all together
 *       and attach this object to the users list handles.
 *----------------------------------------------------------------------*/
{
	listclass_t listclass;
	_embedh_t newembedlist;
	_embedh_t oldembedlist;
    list_t    oldlist = *poldlist;
    list_t    newlist;

    LIBASSERT (_LIST_ListCheck(oldlist));
    LIBASSERT (newclass != NULL);

    if (LIST_Count(oldlist) != 0)
    {
        errno = LIST_ERROR_NOTEMPTY;
        return NULL;
    }

	if (LIST_Embedded(oldlist))
	{
		oldembedlist = _LIST_EmbedOuter(oldlist);
		oldlist = oldembedlist->self;
	}
	else
	{
		listclass = oldlist->listclass;
		oldembedlist = (_embedh_t)LIST_Create (LIST_CLASS_EMBED,
												listclass->sizeuser, 
												listclass->initnode,
												listclass->killnode,
												listclass->compare);
		if (oldembedlist == NULL)
		{
			return NULL;			
		}
		oldembedlist->self = oldlist;
		oldembedlist->listclass->sizenode = oldlist->listclass->sizenode;
		*poldlist = (list_t)oldembedlist;		/* Change users copy */
	}
	/*
	 *	Create the new list entry
	 */
	listclass = oldlist->listclass;
	newlist = LIST_Create ( newclass,
						    listclass->sizeuser,
						    listclass->initnode,
						    listclass->killnode, 
						    compare);
	if (newlist == NULL)
	{
		return NULL;
	}
	newlist->listclass->sizenode += listclass->sizenode;
	/*
	 *	Create the 'embedding' entry
	 */
    listclass = newlist->listclass;
	newembedlist = (_embedh_t)LIST_Create (LIST_CLASS_EMBED,
											listclass->sizeuser, 
											listclass->initnode,
											listclass->killnode,
											listclass->compare);
	if (newembedlist == NULL)
	{
		(void)LIST_Destroy(&newlist);
		return NULL;			
	}
	newembedlist->self = newlist;
	newembedlist->listclass->sizenode = newlist->listclass->sizenode;

	/*
	 *	Now link the two lists
	 */
	oldembedlist->outer = newembedlist;
	newembedlist->inner = oldembedlist;
	
	return (list_t)newembedlist;
}



/*========================================================================
 *
 *  This is the list of vectored functions
 *  that are used to support this list type.
 *
 *------------------------------------------------------------------------*/

static struct _listmaster master =
        {
        &_LIST_EmbedCount,
        &_LIST_EmbedFirst,
        &_LIST_EmbedLast,
        &_LIST_EmbedNext,
        &_LIST_EmbedPrevious,
        &EMBED_Clone,
        &EMBED_Destroy,
        &_LIST_EmbedNew,
        &_LIST_EmbedAdd,
        &_LIST_EmbedAfter,
        &_LIST_EmbedRemove,
        &_LIST_EmbedFind,
        LIST_TYPE_EMBEDDED,
#ifdef DEBUG
        &_LISTembed,
        "Embedded list"
#endif /* DEBUG */
        };

/*
 *  This is the default settings for a new
 *  object of this type.
 */
struct _listclass _LISTembed =
        {
        &master,                    /* Pointer to master class */
        sizeof(struct _embedh),     /* Size of the list header */
        0,							/* Size of the node header (dummy) */
        0,                          /* Clone count */
        0,                          /* Size of node data */
        NULL,                       /* User node Initialise */
        NULL,                       /* User node Destroy */
        NULL,                       /* user node Compare */
        };
