/*
 *      h a s h _ c
 *
 *  This contains routines that are specific
 *  to the implementation of hash tables
 *
 *
 * (c) Copyright 1996-1997 David J. Walker
 *     Email:  d.j.walker@x400.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  12 May 96   DJW   - First version
 */

#include <liblist.h>
#include "liblistx.h"
#include <stdlib.h>
#include <string.h>

/*=======================================================================
 *
 *      The following structure is used for by the default
 *      hash table routines supplied with this library
 *----------------------------------------------------------------------*/

struct mynode {
        char *  name;
        };
typedef struct mynode * MYNODE_t;


/*===================================================== HASH_INITNODE */
static  
int		HASH_InitNode (node_t node, va_list ap)
/*		~~~~~~~~~~~~~
 *	This is the default node initalisation routine used
 *	if no alternative one is supplied.
 *-----------------------------------------------------------------------*/
{
	char * name = va_arg (ap, char *);

	if ((((MYNODE_t)node)->name = malloc (strlen (name) + 1)) == NULL)
	{
		return LIST_ERROR_NOMEMORY;
	}
	(void)strcpy (((MYNODE_t)node)->name, name);
	return LIST_ERROR_NONE;
}

/*===================================================== HASH_KILLNODE */
static 
int		HASH_KillNode (node_t node)
/*		~~~~~~~~~~~~~
 *	This is the default node destruction routine used
 *	if no alternative one is supplied.
 *-----------------------------------------------------------------------*/
{
	free (((MYNODE_t)node)->name);
	return LIST_ERROR_NONE;
}

/*===================================================== HASH_COMPNODE */
static 
int		HASH_CompNode (node_t node1, node_t node2)
/*		~~~~~~~~~~~~~
 *	This is the default node comparison routine used
 *	if no alternative one is supplied.
 *-----------------------------------------------------------------------*/
{
	return strcmp (((MYNODE_t)node1)->name, ((MYNODE_t)node2)->name);
}



/*===================================================== HASH_DEFAULTCALC */
static
size_t	HASH_DefaultCalc (char *name)
/*		~~~~~~~~~~~~~~~~
 *	This is the default hash calculation algorithm used
 *	if no alternative one is supplied.
 *-----------------------------------------------------------------------*/
{
    register size_t sum;

    for (sum = 0 ; *name ; name++ )
    {
        sum += (sum << 5) + (size_t)*name;
    }
    return (sum);
}

/*============================================================= HASH_CALC */
extern  
size_t  HASH_NodeCalc (node_t node)
/*      ~~~~~~~~~~~~~
 *  This is the default hash table routine used if the user supplies
 *  no alternative  (it is the same as used in the C68 compiler).
 *------------------------------------------------------------------------*/
{
	return HASH_DefaultCalc (((MYNODE_t)node)->name);
}


/*====================================================== HASH_FINDCALC */
extern
size_t  HASH_FindCalc (va_list ap)
/*      ~~~~~~~~~~~~~
 *  This is the default hash algorithm used for LIST_Find
 *  methods on Hash Table nodes.
 *---------------------------------------------------------------------*/
{
    return HASH_DefaultCalc (va_arg(ap, char *));
}



/*============================================================ HASH_CLONE */
static
list_t  HASH_Clone  (list_t list)
/*      ~~~~~~~~~~
 *	This is the hash table clone routine.
 *	Fundementally we simply clone the given
 *	list and the attached list if entry, copying
 *	across any relevant infromation.
 *------------------------------------------------------------------------*/
{
	_hashh_t    oldhashlist = (_hashh_t)list;
	_hashh_t	newhashlist;
	list_t		newclonelist;

	/*
	 *	If there was an attached list on the
	 *	old one, then clone that entry.
	 */
	if (oldhashlist->clonelist == NULL)
	{
		newclonelist = NULL;
	}
	else
	{
		newclonelist = _LIST_CloneList (oldhashlist->clonelist);
		if (newclonelist == NULL)
		{
			return NULL;
		}
	}
	/*
	 *	Now try and clone the main list
	 */
	newhashlist = (_hashh_t)_LIST_CloneList ((list_t)oldhashlist);
	if (newhashlist == NULL)
	{
		(void)newclonelist->listclass->master->destroy (&newclonelist);
	}
	/*
	 *	All memory allocations successful.
	 *	finish setup of data in header.
	 */
	newhashlist->clonelist = newclonelist;
	newhashlist->bucketcount = oldhashlist->bucketcount;
	newhashlist->nodehash = oldhashlist->nodehash;
	newhashlist->findhash = oldhashlist->findhash;

	return (list_t)newhashlist;
}


/*========================================================== HASH_NEWNODE */
static
node_t  HASH_NodeNew  (list_t list, va_list ap)
/*      ~~~~~~~~~~~~
 *  Handle the setting up of a new node.
 *	If there is no attached list then we
 *	have to handle it at this level, and if
 *	not we simply pass it down a level.
 *------------------------------------------------------------------------*/
{
	register _hashh_t hashlist = (_hashh_t)list;

	/*
	 *	Check if list set up yet.
     *  If not then try and set it up.
	 */
	if (hashlist->buckets == NULL)
	{
		int reply = LIST_HashSetup (list, 0, NULL, NULL, NULL);

		if (reply != LIST_ERROR_NONE)
		{
			errno = reply;
		}
		return NULL;
	}

	if (hashlist->clonelist == NULL)
	{
		/*
		 *	Set up new node entry for the case in
		 *	which there is not attached list.
		 */
		return _LIST_NewNode ((list_t)hashlist, ap);
	}

	/*
	 *	Set up a new emtry by passing the
	 *	requirement down a level.
	 */
	return hashlist->clonelist->listclass->master->nodenew(hashlist->clonelist,ap);
}


/*======================================================== LIST_HASHSETUP */
extern
int     LIST_HashSetup    (list_t list,
/*      ~~~~~~~~~~~~~~*/   size_t  bucketcount,
						   size_t (*nodehash)(node_t),
                           size_t (*findhash)(va_list),
                           list_t linkedlist)
/*
 *	This function must be called once a hash table
 *	has been created to finish the setup details.
 *------------------------------------------------------------------------*/
{
    _hashh_t    hashlist = (_hashh_t)list;
    list_t      clonelist;

    LIBDEBUG ("LIST_HashSetup", \
          "'list' parameter is not a valid list instance", \
          _LIST_ListCheck(list))

#ifdef DEBUG
	if (linkedlist != NULL && linkedlist != LIST_HASHONLY)
	{
    LIBDEBUG ("LIST_HashSetup", \
          "'linkedlist' parameter is not a valid list instance", \
          _LIST_ListCheck(linkedlist))

	}
#endif /* DEBUG */

	/*
	 *	See if this is first call?
	 */
	if (hashlist->bucketcount == 0)
	{
		listclass_t cloneclass;
		/*
		 *	YES - set default values
		 */
		if ((hashlist->clonelist = _LIST_CloneClass (LIST_CLASS_LIFO)) == NULL)
		{
			return errno;
		}
		hashlist->bucketcount = 127;
		hashlist->nodehash = HASH_NodeCalc;
		hashlist->findhash = HASH_FindCalc;
		cloneclass = hashlist->clonelist->listclass;
		cloneclass->initnode = HASH_InitNode;
		cloneclass->killnode = HASH_KillNode;
		cloneclass->compare  = HASH_CompNode;
	}
	else
	{
	    /*
		 *  NO - Check that the list is still empty
	     */
	    if (LIST_First((list_t)hashlist) != NULL)
		{
			return EEXIST;
		}
	}
    /*
     *  See if any of the node hashing
     *  algorithms have changed.  Note
	 *	that they must change together.
     */
	if ((nodehash != NULL || findhash != NULL)
	&&  (nodehash == NULL || findhash == NULL))
	{
		return LIST_ERROR_BADHASH;
	}
    if (nodehash != NULL)
    {
        hashlist->nodehash = nodehash;
    }
    if (findhash != NULL)
    {
        hashlist->findhash = findhash;
    }

    /*
     *  See if the bucket details have changed.
     *  If so, destroy the old bucket contents,
     *  and then instantiate the new ones.
     */
    if (   bucketcount != hashlist->bucketcount
		|| linkedlist != hashlist->clonelist
        || (linkedlist == LIST_HASHONLY & hashlist->clonelist != NULL)
        || (linkedlist != LIST_HASHONLY & hashlist->clonelist == NULL)
        || linkedlist->listclass != hashlist->clonelist->listclass 
		)
    {
		/*
		 *	See if size of hash list specified.
		 */
        if (bucketcount != 0)
        {
            hashlist->bucketcount = bucketcount;
        }
		/*
		 *	See if linked list is being changed
		 */
        if (linkedlist != NULL)
        {
			clonelist = hashlist->clonelist;
			if (clonelist != NULL)
			{
				(void) clonelist->listclass->master->destroy (&clonelist);
			}
			if (linkedlist == LIST_HASHONLY)
			{
				hashlist->clonelist = NULL;
			}
			else
			{
				if ((hashlist->clonelist = _LIST_CloneList(linkedlist)) == NULL)
				{
					return errno;
				}
			}
		}
		/*
		 *	If we have already set up a list
		 *	then free it down.
		 */
		if (hashlist->buckets != NULL)
		{
	        free (hashlist->buckets);
		    hashlist->buckets = NULL;
		}
	}
	/*
	 *	Now set up a new hash table
	 *	Note that we do not try to actually
	 *	assign linked lists to nodes until
	 *	the point at which we try to first
	 *	add data to that node.
	 */
	if (hashlist->buckets == NULL)
	{
        hashlist->buckets = calloc (hashlist->bucketcount, sizeof(struct _hash));
        if (hashlist->buckets == NULL) {
            return LIST_ERROR_NOMEMORY;
        }
    }
	/*
	 *	See if there is an attached list that needs to
	 *	inherit any functions from the hash list level.
	 */
	if (hashlist->clonelist != NULL)
	{
		listclass_t  hashclass = hashlist->listclass;
		listclass_t  cloneclass = hashlist->clonelist->listclass;

		if (hashclass->initnode != NULL 
		 || hashclass->killnode != NULL
		 || hashclass->compare != NULL)
		{
			cloneclass->initnode = hashclass->initnode;
			cloneclass->killnode = hashclass->killnode;
			cloneclass->compare = hashclass->compare;
		}
		cloneclass->sizeuser = hashclass->sizeuser;
	}
	/*
	 *	Finally ensure that the 'sizenode' field is set
	 *	up to expedite calculating node offsets later.
	 */
	for (hashlist->listclass->sizenode = 0, clonelist = hashlist->clonelist ;
				clonelist != NULL ; 
						clonelist = ((_hashh_t)clonelist)->clonelist)
	{
		hashlist->listclass->sizenode = clonelist->listclass->sizenode;
		if (LIST_Type(clonelist) != LIST_TYPE_HASH)
		{
			break;
		}
	}

    return LIST_ERROR_NONE;
}


/*========================================================================
 *
 *  This is the list of vectored functions
 *  that are used to support this list type.
 *

 *------------------------------------------------------------------------*/

static struct _listmaster master =
        {
        &_LIST_ListCount,
        &_LIST_HashFirst,
        &_LIST_HashLast,
        &_LIST_HashNext,
        &_LIST_HashPrevious,
        &HASH_Clone,
        &_LIST_Destroy,
        &HASH_NodeNew,
        &_LIST_HashAdd,
        &_LIST_HashAfter,
        &_LIST_HashRemove,
        &_LIST_HashFind,
        LIST_TYPE_HASH,
#ifdef DEBUG
        &_LISThash,
        "Hash Table"
#endif /* DEBUG */
        };

/*
 *  This is the default settings for a new
 *  object of this type.
 */
struct _listclass _LISThash =
        {
        &master,                    /* Pointer to master class data */
        sizeof(struct _hashh),      /* Size of the list header */
        0,							/* Size of the node header */
        0,                          /* Size of user node data */
        0,                          /* Clone count */
        NULL,						/* User node Initialise */
        NULL,						/* User node Destroy */
        NULL,						/* User node Compare */
        };


