/*
 *      l i s t t e s t
 *      ~~~~~~~~~~~~~~~
 *  This is a test program for the LIBLIST library.
 *  It can also be considered an example of how to
 *  use the various methods and properties.
 *
 *
 * (c) Copyright 1996 David J. Walker
 *     Email:  d.j.walker@slh0101.wins.icl.co.uk
 *
 *     Permission to copy and/or distribute granted under the
 *     following conditions:
 *
 *     1). This notice must remain intact.
 *     2). The author is not responsible for the consequences of use
 *         this software, no matter how awful, even if they
 *         arise from defects in it.
 *     3). Altered version must not be represented as being the
 *         original software.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  25 May 96   DJW   - First version
 *
 *  28 Mar 98   DJW   - Added some additional tests to make sure some error
 *                      conditions are properly tested.
 */


#include <liblist.h>
#include "liblistx.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef QDOS
#include <qdos.h>
#include <unistd.h>
#endif /* QDOS */

#ifdef DEBUG
char    _prog_name[] = "LibTest_DBG";
#else
char    _prog_name[] = "LibTest";
#endif /* DEBUG */

int	online;
/*
 *  A definition for the user part of the list nodes
 *  
 *  Note that using a 'typedef' to create a pointer
 *  to this structure type will make later code
 *  with casts much cleaner and easier to read.
 */
typedef struct MYNODE {
        char  * name;
        time_t  when;
    } * MYNODE_t ;



/*============================================================== MYEXIT */
void   myexit(int code)
/*
 *	Exit, but if on-line give a chance to read message.
 *----------------------------------------------------------------------*/
{
    if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}
	exit (code);
}



/*=========================================================== INITNODE */

int     initnode (node_t node, va_list ap)
/*
 *  A 'Node initialisation' function.
 *
 *  It is passed the following parameters via 
 *  the 'ap' variable parameter pointer.
 *      char * name;
 *
 *  The function simply uses the strdup() function to 
 *  allocate the memory to hold a copy of the name and
 *  copy it across.  A pointer to the resulting string
 *  is stored in the node.  This can later be used for
 *  comparison and searching purposese.
 *---------------------------------------------------------------------*/
{
    struct MYNODE *mynode;
    char * name;

   /*
    *   To avoid us having to use excessive numbers of casts,
    *   set up a pointer to the user part of the data that
    *   has the correct data type to enable us to access the
    *   structure members using it.
    *
    *   Also get the 'name' parameter from the variable part
    *   of the parameter list.
    */
    mynode = (MYNODE_t)node; 
    name = va_arg(ap,char *); 

    /*
     *  This next part would probably not
     *  normally be present in most programs.
     *  However as this a test program, we
     *  have far more consistency checks
     *  and progress reports than would
     *  normally be the case.
     */

    (void)printf ("  initialising node at $%p (%s)", node, name);

    if (strlen(name) < 10)
    {
        (void)printf (" *** name parameter invalid ($%p) ***\n",name);
        myexit(-1);
    }

	/*
	 *	Get memory to hold name.
	 *  Could use strdup(), but this is not
	 *  an ANSI function, and thus is not
	 *  available on all systems.
	 */
	mynode->name = malloc (strlen(name)+1);

    /*
     *  Now some additional checks and reports.
     *  Most user programs assume that strdup()
     *  always works although obviously they
     *  probably should not!
     */
    if (mynode->name == NULL)
    {
        (void)printf (" *** malloc() failed (errno=%d) ***\n",errno);
        myexit (ENOMEM);
    }
	(void)strcpy(mynode->name, name);
    (void)printf ("  (name at $%p)  DONE\n", mynode->name);

    /*
     *  Finally make a good return
     */
    return 0;
}

/*============================================================== KILLNODE */

int     killnode (node_t node)
/*
 *  A 'Node destruction' function.
 *
 *  It must free the memory that is used
 *  to hold the name information.
 *------------------------------------------------------------------------*/
{
    (void) printf ("  killing node at $%p (name=$%p,'%s')", 
                        node, ((MYNODE_t)node)->name, ((MYNODE_t)node)->name);
    free (((MYNODE_t)node)->name);

    (void) printf ("   DONE\n");
    return (0);
}

/*============================================================== COMPARE */

int     compare (node_t node1, node_t node2)
/*  
 *  A 'Node compare' function.
 *
 *  It compares on the name fields
 *  so that we can order a list 
 *  alphabetically be node name.
 *------------------------------------------------------------------------*/
{
    return strcmp (((MYNODE_t)node1)->name, 
                   ((MYNODE_t)node2)->name);
}

/*============================================================ FINDNODE */

int   findnode (node_t node, va_list ap)
/*
 *  A 'Node find' function.
 *
 *  It is passed the following parameters via 
 *  the 'ap' variable parameter pointer.
 *      char * name;
 *----------------------------------------------------------------------*/
{
    char * name;

    name = va_arg(ap, char *);

    return strcmp(((struct MYNODE *)node)->name, name);
}

/*============================================================ ENUMNODE */

long   enumnode (node_t node, va_list ap)
/*
 *  A 'Node enumeration' function.
 */
{
    struct MYNODE * mynode;

    (void)ap;
    mynode = (MYNODE_t)node;
    (void)printf ("  Node at $%p, name='%s'\n", node, mynode->name);
    return (0);
}

/*====================================================================*/

/*
 *	Data used in basic functional tests
 */
static  char * nodenames[] = {
            "always first",
            "z 2nd unordered",
            "third unordered",
            "fourth unordered",
            "fifth unordered",
            "sixth unordered",
            "seventh unordered",
            "eight unordered",
            "ninth unordered",
            "tenth (last) unordered",
            NULL        /* End of list */
            };

/*
 *	data used in more comprehensive tests
 */
static char * lots[] = {
        "hypertext",
        "Addressing",
        "X11",
        "Tools",
        "Proposal.wn",
        "Protocols",
        "NeXT",
        "Daemon",
        "Test",
        "Administration",
        "LineMode",
        "DesignIssues",
        "MarkUp",
        "Macintosh",
        "Proposal.rtf.wn",
        "FIND",
        "Paper",
        "Tcl",
        "Talks",
        "Architecture",
        "VMSHelp",
        "Provider",
        "Archive",
        "SLAC",
        "Project",
        "News",
        "Viola",
        "Users",
        "FAQ",
        "WorkingNotes",
        "Windows",
        "FineWWW",
        "Frame",
        "XMosaic",
        "People",
        "All",
        "Curses",
        "Erwise",
        "Carl",
        "MidasWWW",
        "XPM",
        "MailRobot",
        "Illustrations",
        "VMClient",
        "XPA",
        "Clients.html",
        "Library",
        "CERNLIB_Distribution",
        "libHTML",
        "WindowsPC",
        "tkWWW",
        "tk2.3",
        "CVS-RCS",
        "DecnetSockets",
        "SGMLStream",
        "NextStep",
        "CVSRepository_old",
        "ArthurSecret",
        "CVSROOT",
        "HytelnetGate",
        "cern.www.new.src",
        "Conditions",
        "HTMLGate",
        "Makefile",
        "Newsgroups.html",
        "People.html",
        "Bugs.html",
        "Summary.html",
        "zDesignIssues.wn",
        "HT.draw",
        "HTandCERN.wn",
        "Ideas.wn",
        "MarkUp.wn",
        "Proposal.html",
        "SearchPanel.draw",
        "Comments.wn",
        "Xanadu.html",
        "Storinglinks.html",
        "TheW3Book.html",
        "Talk_Feb-91.html",
        "JFosterEntry.txt",
        "Summary.txt",
        "Bibliography.html",
        "HTandCern.txt",
        "Talk.draw",
        "zDesignNotes.html",
        "Link.html",
        "Status.html",
        "http.txt",
        "People.html~",
        "TAGS",
        "summary.txt",
        "Technical.html",
        "Terms.html",
        "JANETAccess.html",
        "People.txt",
        "README.txt",
        "CodingStandards.html",
        "Copyright.txt",
        "Status_old.html",
        "patches~",
        "RelatedProducts.html",
        "Implementation",
        "History.html",
        "Makefile.bak",
        "Makefile.old",
        "Policy.html",
        "WhatIs.html",
        "TheProject.html",
        "Notation.html",
        "Helping.html",
        "Cyber-WWW.sit.Hqx",
        "Glossary.html",
        "maketags.html",
        "IntroCS.html",
        "Contrib",
        "Help.html",
        "CodeManagExec",
        "HT-0.1draz",
        "Cello",
        "TOPUB",
        "BUILD",
        "BUILDALL",
        "Lynx",
        "ArthurLibrary",
        "RashtyClient",
        "#History.html#",
        "PerlServers",
        "modules",
        "NCSA_httpd",
        "MAIL2HTML",
        "core",
        "EmacsWWW"
        };


/*========================================================== CREATE_LIST */

list_t  create_list (  listclass_t  listclass,
                       int  (*initfunc)(node_t, va_list),
                       int  (*killfunc)(node_t),
                       int  (*compfunc)(node_t, node_t),
                       char * description)
/*
 *  This function simply creates a list according to the
 *  various parameters passed in, and then checks that
 *  it worked correctly. 
 *  On success the list handle is returned.
 *------------------------------------------------------------------------*/
{
    list_t  list;

    (void)printf ("Creating new list (%s %s)\n", 
                            description, listclass->master->listdesc);
    list = LIST_Create(listclass, sizeof (struct MYNODE),
                        initfunc, killfunc, compfunc);
    if (list == NULL)
    {
        (void)printf ("Failed to create list (%s %s).  errno=%d\n",
                        description, listclass->master->listdesc, errno);
        myexit (-1);
    }
    (void)printf ("list created (%s %s).  Pointer = $%p\n\n", 
                            description, listclass->master->listdesc, list);
    return (list);
}


/*=========================================================== CLONE_LIST */

list_t  clone_list  (list_t  master,
                     char *  description)
/*
 *  Clone a given list and check that it worked.
 *  On success return new list handle.
 *-----------------------------------------------------------------------*/
{
    list_t  list;

    list = LIST_Clone (master);
    if (list == NULL)
    {
        (void)printf ("Failed to clone list (%s %s).  errno=%d\n",
                        description, list->listclass->master->listdesc, errno);
        myexit (-1);
    }
    (void)printf ("list cloned (%s %s).  Pointer = $%p\n\n", 
                            description, list->listclass->master->listdesc, list);
    return (list);
}


/*=========================================================== EMBED_LIST */

list_t  embed_list  (listclass_t listclass,
                     list_t *  master,
                     int (*  compfunc)(node_t,node_t),
                     char *  description)
/*
 *  Clone a given list and check that it worked.
 *  On success return new list handle.
 *-----------------------------------------------------------------------*/
{
    list_t  list;

    list = LIST_Embed (listclass,master,compfunc);
    if (list == NULL)
    {
        (void)printf ("Failed to embed list (%s %s).  errno=%d\n",
                        description, list->listclass->master->listdesc, errno);
        myexit (-1);
    }
    (void)printf ("list embedded (%s %s) in list at %p.  Pointer = $%p\n\n", 
                    description, list->listclass->master->listdesc, master, list);
    return (list);
}


/*============================================================ ADD_LIST */

void    add_list (list_t  list,
                  char * description)
/*
 *  This function adds data nodes to a list.
 *------------------------------------------------------------------------*/
{
    char ** pptr;

    /*
     *  Now we add our list of nodes
     *  to this list.
     */
    (void)printf ("adding nodes(%s %s)\n",
                        description, list->listclass->master->listdesc);
    for (pptr = nodenames; *pptr != NULL ; pptr++)
    {
        if (LIST_NewAdd(list, *pptr) == NULL)
        {
            (void)printf ("Failed to add a new node.  errno=%d\n",errno);
            myexit (-1);
        }
    }
    (void)printf ("list now populated (%s %s).\n  Count = %ld, Ordered=0x%d\n", 
                            description, list->listclass->master->listdesc,
                            LIST_Count(list), LIST_Ordered(list));
    if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}
    return;
}


/*============================================================ TEST_LIST */

void    test_list ( list_t  list,
                    char * description)
/*
 *  This function runs a test of one particular list.
 *  It is called repeatedly to test each list type in turn.
 *------------------------------------------------------------------------*/
{
    node_t  node;
    long    lreply;

    (void)printf ("Enumerating list (%s %s)\n", 
                        description, list->listclass->master->listdesc);
    lreply = LIST_Enumerate(list,&enumnode);
    if (lreply != 0)
    {
        (void)printf ("Failed while enumerating list.  reply=%ld\n",lreply);
        myexit (-1);
    }
    (void)printf ("list enumeration completed\n");
    if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}

    /*
     *  We now do various calls on our new
     *  list to check they all give the
     *  expected result.  The expected result
     *  needs to be checked against the
     *  enumeration list.
     */
    (void)printf ("testing miscellaneous calls (%s %s)\n", 
                        description, list->listclass->master->listdesc);
    node = LIST_First(list);
    (void)printf ("LIST_First(): First node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    node = LIST_Last(list);
    (void)printf ("LIST_Last(): Last node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
	node = LIST_Find(list, findnode, "Non-existent node");
	if (node != NULL)
	{
		(void)printf ("Find of non-existent node worked!, errno=%d\n", errno);
		myexit (-1);
	}
    node = LIST_Find(list, findnode, nodenames[3]);
    if (node == NULL)
    {
        (void)printf ("Find failed.  errno=%d\n", errno);
        myexit (-1);
    }
    (void)printf ("LIST_Find(): Node found at $%p (%s)\n",
                                            node, ((MYNODE_t)node)->name);
    lreply = LIST_Index(list,node);
    (void)printf ("LIST_Index(): Node list position is %ld\n", lreply);
    node = LIST_Previous(list,node);
    (void)printf ("LIST_Previous(): Previous node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    node = LIST_Position(list,lreply);
    (void)printf ("LIST_Position(%ld):  node at $%p (%s)\n",
                                lreply, node, ((MYNODE_t)node)->name);
    node = LIST_Next(list,node);
    (void)printf ("LIST_Next(): Next node at $%p (%s)\n",
                                node, ((MYNODE_t)node)->name);
    if (online) {(void)printf ("Press ENTER to continue\n");  (void)getchar();}
    return;
}


/*=========================================================== DESTROY_LIST */

void    destroy_list  (list_t * plist,
                       char *   description)
/*
 *-------------------------------------------------------------------------*/
{
    int     ireply;

    (void)printf("Destroying list at %p (%s)\n", 
                            *plist, description);
    ireply = LIST_Destroy(plist);
    if (ireply != 0)
    {
        (void)printf ("Failure while destroying list at address $%p.  reply=%d\n",
                                                *plist, ireply);
    }
	else
	{
		(void)printf ("destroyed list at $%p OK\n", *plist);
    }
	return;
}

/*
 *	List types that need to be tested
 */
   struct _testdata {
        struct _listclass *  listclass;
        int (*initnode)(node_t,va_list);
        int (*killnode)(node_t);
        int (*compnode)(node_t, node_t);
        char *  desc;
    } listtypes[] = {
        {LIST_CLASS_SINGLE, &initnode, &killnode, &compare, "SINGLE ORDERED"},
        {LIST_CLASS_SINGLE, &initnode, &killnode, NULL, "SINGLE unordered"},
        {LIST_CLASS_DOUBLE, &initnode, &killnode, &compare, "DOUBLE ORDERED"},
        {LIST_CLASS_DOUBLE, &initnode, &killnode, NULL, "DOUBLE unordered"},
        {LIST_CLASS_FIFO,  &initnode, &killnode, NULL, "FIFO unordered"},
        {LIST_CLASS_LIFO,  &initnode, &killnode, NULL, "LIFO unordered"},
        {LIST_CLASS_BTREE, &initnode, &killnode, &compare, "BTREE ORDERED"},
        {LIST_CLASS_BTREE, &initnode, &killnode, NULL, "BTREE unordered"},
        {LIST_CLASS_BALTREE, &initnode, &killnode, &compare, "BALTREE ORDERED"},
        {LIST_CLASS_BALTREE, &initnode, &killnode, NULL, "BALTREE unordered"},
        {NULL,NULL,NULL,NULL,NULL}
        };


/*================================================================ MAIN */

int     main (int argc, char * argv[])
{
 
    struct _testdata *ptr1;
    struct _testdata *ptr2;
    struct _testdata *ptr3;
    list_t  master;
    list_t  list;
    list_t  list2;

    (void) argc;    /* lets stop compiler warnings */
    (void) argv;    /* about variables not used    */
    (void) lots;

    online = (argc < 2);

    (void)printf ("TEST OF LIBLIST LIBRARY\n");
    (void)printf ("~~~~~~~~~~~~~~~~~~~~~~~\n\n");

    for (ptr1 = listtypes ; ptr1->listclass != NULL ; ptr1++)
    {
        char * typedesc;

        (void)printf ("\n---------- STARTING TEST (%s) ----------\n\n", ptr1->desc);
        master = create_list (ptr1->listclass,
                              ptr1->initnode,
                              ptr1->killnode,
                              ptr1->compnode,
                              ptr1->desc);
        typedesc = master->listclass->master->listdesc;

        list = clone_list (master,
                           ptr1->desc);

        add_list  (list,
                   ptr1->desc);

        test_list (list, 
                   ptr1->desc);

        destroy_list (&list,
                      ptr1->desc);                      

        destroy_list (&master,
                      ptr1->desc);                      

        (void)printf ("\n---------- COMPLETED tests (%s %s) ----------\n\n",
                                    ptr1->desc, typedesc);
        if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}
    }

    /*
     *  Now do some special tests for embedded tables
     */
    (void)printf ("\n========== STARTING TEST of EMBEDDED tables ==========\n\n");
    ptr1 = &listtypes[4];
    master = create_list (ptr1->listclass,
                          ptr1->initnode,
                          ptr1->killnode,
                          ptr1->compnode,
                          ptr1->desc);

    ptr2 = &listtypes[5];
    list = embed_list (ptr2->listclass,
                       &master,
                       ptr2->compnode,
                       ptr2->desc);

    ptr3 = &listtypes[0];
    list2 = embed_list (ptr3->listclass,
                       &master,
                       ptr3->compnode,
                       ptr3->desc);
    add_list  (list,
               ptr2->desc);

    test_list (master, 
               ptr1->desc);

    test_list (list, 
               ptr2->desc);

    test_list (list2, 
               ptr3->desc);

    destroy_list (&list,
                  ptr2->desc);                      
#if 0
	destroy_list (&master,
		           ptr1->desc);

	destroy_list (&list2,
		          ptr3->desc);
#endif

    (void)printf ("\n--------- COMPLETED EMBEDDED list tests ----------\n");
    if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}

    /*
     *  Now do some special tests for hash tables
     */
    (void)printf ("\n========== STARTING TEST of HASH tables ==========\n\n");
    list = create_list (LIST_CLASS_HASH,
                        initnode,
                        killnode,
                        compare,
                        "Hash Table");

    (void)LIST_HashSetup (list,
                          (sizeof (lots) / sizeof (char *)) + 5,
                          NULL,
                          NULL,
                          NULL);

    add_list  (list,
               "Hash Table");

    test_list (list, 
               "Hash Table");

    destroy_list (&list,
                  "Hash Table");

    (void)printf ("\n--------- COMPLETED HASH table tests ----------\n");
    if (online) {(void)printf ("Press ENTER to continue\n"); (void)getchar();}

    /*
     *  We seem to have finally finished
     */
    (void)printf ("\nALL TESTS COMPLETED OK");
    (void)printf ("\n~~~~~~~~~~~~~~~~~~~~~~\n\n");
    return (0);
}

